/*
 * XFormat (xf) (C) 1997 by Karsten Ballder <Ballueder@usa.net>
 *
 * This program is protected and licensed under the terms and
 * conditions of the GNU General Public License, Version 2.
 *
 */
#include	"config.h"	// using configure

#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	<signal.h>

#define _NO_USES_DEFINES
#include <wx.h>


#include	"xformat.xpm"
#include	"wxtab.h"
#include	"wxExtApp.h"
#include	"appconf.h"

#define VERSION	"1.01"
#define	APPNAME	"xf"

#ifdef	HAVE_LIBINTL
#	include	<libintl.h>
#	define _(String) gettext (String)
#	define	PACKAGE	APPNAME
#else
#	define _(String) String
#endif


#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/stat.h>
extern "C" 
{
#include <linux/fd.h>
#include <linux/fs.h>
}
#include <sys/ioctl.h>


/* fdformat.c  -  Low-level formats a floppy disk. */


static int ctrl;
struct floppy_struct param;
#define FLOPPY_MAJOR 2
#define SECTOR_SIZE 512
//#define PERROR(msg) { perror(msg); exit(1); }
#define PERROR(msg) { xfframe->ErrMsg(msg); }
#define	MSG(msg)	xfframe->SetStatusText(msg);

char	*lang[] =
{	"default", "English", "Deutsch", "Francais", "Italiano", NULL };

char	*reallang[] =
{	NULL, "en", "de", "fr", "it_IT", NULL };

char	*drivenames[] =
{	"fd0","fd1", NULL };

char 	*capacitynames[] =
{
   "360 KB","720 KB","1.2 MB","1.44 MB","1.76 MB",
   NULL
};

char	*realcapacitynames[] =
{	"360","720","1200","1440","1476",NULL };

char	*filesystems[] =
{	"msdos", "ext2", "minix", "xiafs", "ext", NULL };

class xfApp : public wxExtApp
{
   wxIcon	*icon;
public:
   xfApp():wxExtApp(APPNAME) {}
   wxFrame	*OnInit(void);
   void	do_mkfs(int tracks, int secs);
};

// entries for configuration file:
#define	OPT_DEV	"Device"
#define OPT_MKFS "MakeFileSystem"
#define	OPT_FS	"FileSystem"
#define	OPT_CAP	"Capacity"
#define	OPT_VERIFY	"Verify"
#define	OPT_LANG	"Language"
#define	OPT_POSITION	"RememberPosition"

// defaults:
#define	OPT_DEV_D	"0"
#define OPT_MKFS_D	"1"
#define	OPT_FS_D	"0"
#define	OPT_CAP_D	"3"
#define	OPT_VERIFY_D	"1"
#define	OPT_LANG_D	"0"
#define OPT_POSITION_D	"0"

enum
{	MENU_FILE, MENU_FILE_EXIT,
	MENU_FILE_FORMAT, MENU_FILE_ABORT,
	MENU_OPT, MENU_OPT_PREFS,
	MENU_HELP, MENU_HELP_ABOUT,
	OPT_TAB_DEV,
	OPT_TAB_VER,
	OPT_TAB_FS,
	OPT_TAB_LANG,
	OPT_TAB_POS
};


// This statement initialises the whole application
xfApp myapp;

void	abort_button_function(void);

#define FLOPPY_MAJOR 2
#define SECTOR_SIZE 512

class xfFrame : public wxFrame
{
   char buffer[256], name[256], buffer2[256];
   FileConfig *config;
   int blocks;
   
public:
   volatile int	formatting;
   pid_t	child_pid;
   wxTabbedPanel *dialog;
   wxPanel	*panel;
   wxGauge	*verify_gauge, *format_gauge;
   wxChoice	*drive, *capacity, *fs, *language;
   wxCheckBox  *verifyflag, *mkfsflag, *positionflag;
   wxButton	*format_button, *abort_button;
   xfFrame(wxFrame *frame, int x, int y, int w, int h);
   void	ErrMsg(const char *msg);
   void	format(void);
   void	do_mkfs(int tracks, int secs);
   void	abort(void) {formatting = 0;}

   ~xfFrame();
   
   Bool	OnClose(void)
      {
	 rememberPosition();
	 delete this;
	 return FALSE;
      }
   void rememberPosition(void);
   void OnMenuCommand(int id);
   void	OptionsDialog(void);
   void	setConfig(void);
};



void xfFrame::ErrMsg(const char *msg)
{
   sprintf(buffer,_("error: %s (%s)"), msg, strerror(errno));
   wxMessageBox(buffer, _("Error"),wxOK|wxCENTRE|wxICON_EXCLAMATION,this);
   SetStatusText(buffer);
}

void	xfFrame::do_mkfs(int tracks, int secs)
{
   char dev;

   if(atoi(config->readEntry(OPT_MKFS,OPT_MKFS_D)))
   {
      int rc;
      int blocks =
	 atoi(realcapacitynames[atoi(config->readEntry(OPT_CAP,OPT_CAP_D))]);
      
      if(atoi(config->readEntry(OPT_FS,OPT_FS_D)) == 0)
      {
	 dev = 'A' + atoi(config->readEntry(OPT_DEV,OPT_DEV_D));
	  

	 sprintf(buffer,"mformat %c: -t %d -s %d", dev, tracks, secs);
	 SetStatusText(buffer);
      }
      else
      {
	 sprintf(buffer,_("executing \"/sbin/mkfs -t %s %s %d\""),
		 filesystems[atoi(config->readEntry(OPT_FS,OPT_FS_D))], name, blocks);
	 SetStatusText(buffer);
	 sprintf(buffer,_("/sbin/mkfs -t %s %s %d"),
		 filesystems[atoi(config->readEntry(OPT_FS,OPT_FS_D))], name, blocks);
      }
      if((rc = system(buffer)) != 0)
      {
	 sprintf(buffer,_("Command returned error value %d."), rc);
	 SetStatusText(buffer);
      }
      else
	 SetStatusText(_("Created filesystem."));
   }    
}


xfFrame 	*xfframe;


static int format_disk(char *name, int *tracks, int *secs)
{
   struct stat st;
   struct format_descr descr;
   unsigned track;
   char buffer[256];
   int rc = 1; // success
   
   if (stat(name,&st) < 0)
   {
      PERROR(name);
      return 0;
   }
   if (access(name,W_OK) < 0)
   {
      PERROR(name);
      return 0;
   }
   if ((ctrl = open(name,3)) < 0)
   {
      PERROR(name);
      return 0;
   }
#ifdef __alpha__
   if (ioctl(ctrl,FDGETPRM,(long) &param) < 0) 
#else 
      if (ioctl(ctrl,FDGETPRM,(int) &param) < 0)
#endif
      {
	 PERROR(_("Could not determine current format type"));
	 rc = 0;
	 goto endfmt;
      }
   sprintf(buffer,_("%sle-sided, %d tracks, %d sec/track. Total capacity %d kB."),
	   param.head ? _("Doub") :
	   _("Sing"),param.track,param.sect,param.size >> 1);
   MSG(buffer);

   *tracks = param.track;
   *secs = param.sect;

   fflush(stdout);
   if (ioctl(ctrl,FDFMTBEG,NULL) < 0)
   {
      PERROR("ioctl(FDFMTBEG)");
      rc = 0;
      goto endfmt;
   }
   
   xfframe->format_gauge->SetRange(param.track-1);
   xfframe->verify_gauge->SetRange(param.track-1);

   for (track = 0; track < param.track && xfframe->formatting; track++)
   {
      descr.track = track;
      descr.head = 0;
#ifdef __alpha__
      if (ioctl(ctrl,FDFMTTRK,(long) &descr) < 0)
#else
	 if (ioctl(ctrl,FDFMTTRK,(int) &descr) < 0)
#endif
	 {
	    PERROR("ioctl(FDFMTTRK)");
	    rc = 0;
	    goto endfmt;
	 }
      
      xfframe->format_gauge->SetValue(track);
      while(myapp.Pending())
	 myapp.Dispatch();

      if (param.head == 2)
      {
	 descr.head = 1;
#ifdef __alpha__
	 if (ioctl(ctrl,FDFMTTRK,(long) &descr) < 0)
#else
	    if (ioctl(ctrl,FDFMTTRK,(int) &descr) < 0)
#endif
	    {
	       PERROR("ioctl(FDFMTTRK)");
	       rc = 0;
	       goto endfmt;
	    }
      }
   }
   if (ioctl(ctrl,FDFMTEND,NULL) < 0)
   {
      PERROR("ioctl(FDFMTEND)");
      rc = 0;
   }
 endfmt:
   close(ctrl);
   if(rc)
      MSG(_("done"));
   return rc;
}


static int verify_disk(char *name)
{
   unsigned char *data;
   int ctrl,cyl_size,count;
   unsigned cyl;
   char	buffer[256];
   
   cyl_size = param.sect*param.head*512;
   
   if ((data = (unsigned char *) malloc(cyl_size)) == NULL)
   {
      PERROR("malloc");
      return 0;
   }
   MSG("verifying...");
   
   if ((ctrl = open(name,O_RDONLY)) < 0)
   {
      PERROR(name);
      return 0;
   }
   
   xfframe->verify_gauge->SetRange(param.track-1);
   for (cyl = 0; cyl < param.track && xfframe->formatting; cyl++)
   {
      xfframe->verify_gauge->SetValue(cyl);
      while(myapp.Pending())
	 myapp.Dispatch();
      
      if (read(ctrl,data,cyl_size) != cyl_size)
      {
	 PERROR("read()");
	 close(ctrl);
	 return 0;
      }
      for (count = 0; count < cyl_size; count++)
	 if (data[count] != FD_FILL_BYTE) {
	    sprintf(buffer,_("bad data in cyl %d. Continuing ... "),cyl);
	    MSG(buffer);
	    break;
	 }
   }
   MSG("done");
   if (close(ctrl) < 0)
      PERROR(_("close"));
   return 1;
}


void	abort_button_function(void)
{
   xfframe->abort();
}

void	format_button_function(void)
{
   xfframe->format();
   xfframe->format_gauge->SetValue(0);
   xfframe->verify_gauge->SetValue(0);
}

void
xfFrame::format(void)
{
   int tracks, secs, rc;
   
   if(formatting)
      return;
   formatting = 1;

   format_button->Enable(FALSE);
   abort_button->Enable(TRUE);
    
   child_pid = 0;
    
   strcpy(name,_("/dev/"));
   strcat(name,drivenames[atoi(config->readEntry(OPT_DEV,OPT_DEV_D))]);
   strcat(name,"h");
   strcat(name,realcapacitynames[atoi(config->readEntry(OPT_CAP,OPT_CAP_D))]);
    
   rc = format_disk(name, &tracks, &secs);
    
   if(rc && atoi(config->readEntry(OPT_VERIFY,OPT_VERIFY_D)))
      rc = verify_disk(name);

   if(rc && formatting)
      do_mkfs(tracks, secs);
   
   abort_button->Enable(FALSE);
   format_button->Enable(TRUE);

   formatting = 0;
}

xfFrame::xfFrame(wxFrame *frame, int x, int y, int w, int h)
   :wxFrame(frame,_("Xformat"), x,y,w,h)
{
   wxIcon *icon = new wxIcon(xformat_xpm);
   SetIcon(icon);

   wxMenu 	*file_menu, *help_menu, *options_menu;
   wxMenuBar	*menubar;


   config = ((wxExtApp *)wxTheApp)->config;

   file_menu = new wxMenu;
   file_menu->Append(MENU_FILE_FORMAT,_("&Format"));
   file_menu->Append(MENU_FILE_ABORT,_("&Abort"));
   file_menu->AppendSeparator();
   file_menu->Append(MENU_FILE_EXIT,_("&Exit"));

   options_menu = new wxMenu;
   options_menu->Append(MENU_OPT_PREFS,_("&Preferences"));
    
   help_menu = new wxMenu;
   help_menu->Append(MENU_HELP_ABOUT, _("&About"));

    
   menubar = new wxMenuBar;
   menubar->Append(file_menu, _("&File"));
   menubar->Append(options_menu, _("&Options"));
   menubar->Append(help_menu, _("&Help"));

   // Associate the menu bar with the frame
   SetMenuBar(menubar);

   formatting = 0;
   panel = new wxPanel(this);

   format_button = new wxButton(panel,(wxFunction)format_button_function,_("Format"));
   abort_button =  new wxButton(panel,(wxFunction)abort_button_function,_("Abort"));
   abort_button->Enable(FALSE);
    
   panel -> NewLine();

   panel->SetLabelPosition(wxVERTICAL);

   format_gauge = new wxGauge(panel,_("formatted"), 100,
			      -1,-1,-1,40,wxHORIZONTAL);
   verify_gauge = new wxGauge(panel,_("verified"), 100,
			      -1,-1,-1,40,wxHORIZONTAL);

   format_gauge->SetRange(80);
   verify_gauge->SetRange(80);
   format_gauge->SetValue(0);
   verify_gauge->SetValue(0);

   CreateStatusLine(1);
   Show(TRUE);

   OptionsDialog(); // create it but don't show

}

xfFrame::~xfFrame()
{
}

wxFrame *xfApp :: OnInit(void)	/*	main 	*/
{
   int x,y;

   FileConfig *config;
   
   lang[0] = _("default");
   
#ifdef	HAVE_LIBINTL
   setlocale (LC_ALL, "");
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
#endif

   config = ((wxExtApp *)wxTheApp)->config;

   x = atoi(config->readEntry("XPos", "-1"));
   y = atoi(config->readEntry("YPos", "-1"));
   
   xfframe = new xfFrame(NULL, x, y,400,146);

            
   // Associate the menu bar with the frame
   xfframe->Show(TRUE);

   return xfframe;
}



// Intercept menu commands
void xfFrame::OnMenuCommand(int id)
{
   switch (id)
   {
   case MENU_FILE_FORMAT:
      format();
      break;
   case MENU_FILE_ABORT:
      abort();
      break;
   case MENU_FILE_EXIT:
      OnClose();
// delete this;
      break;
   case MENU_OPT_PREFS:
      Fit();
      dialog->Show(TRUE);
      break;
   case MENU_HELP_ABOUT:
   {
      char	buf[4096];
      sprintf(buf,
	 _("XFormat %s, Copyright (C) 1997 by\n"
	   "\n"
	   "Karsten Ballder, 11 (2F3) Yeaman Place,\n"
	   "Edinburgh EH11 1BR, Scotland\n"
	   "(Ballueder@usa.net)\n"
	   "\n"
	   "XFormat has been written using the cross-\n"
	   "platform class library wxWindows (Xt port)\n"
	   "which is freely available under the GNU Library\n"
	   "Public License.\n"
	   "\n"
	   "GUI:\n"
	   "(c) Markus Holzem, Schlossparkstrasse 3,\n"
	   "52072 Aachen, Germany\n"
	   "Goto http://web.ukonline.co.uk/julian.smart/wxwin/ for more\n"
	   "information.\n"),VERSION);
      wxMessageBox(buf, _("XFormat Information"),wxOK,xfframe);
      break;
   }
   }
}

void
xfFrame::rememberPosition(void)
{
   char	buffer[20];
   int x,y;
   
   config->setCurrentPath();
   config->setCurrentPath(APPNAME);
   sprintf(buffer,"%d", positionflag->GetValue());
   config->writeEntry(OPT_POSITION, buffer);
   if(positionflag->GetValue())
   {
      GetPosition(&x,&y);
      sprintf(buffer,"%d",x);
      config->writeEntry("XPos", buffer);
      sprintf(buffer,"%d",y);
      config->writeEntry("YPos", buffer);
   }
   else
   {
      config->deleteEntry("XPos");
      config->deleteEntry("YPos");
   }
}

void
xfFrame::setConfig(void)
{
   char	buffer[20];
   
   // to the global section:
   config->setCurrentPath();
   config->writeEntry(OPT_LANG, reallang[language->GetSelection()]);

   // to the xformat section:
   config->setCurrentPath(APPNAME);
   sprintf(buffer,"%d",drive->GetSelection());
   config->writeEntry(OPT_DEV,buffer);
   sprintf(buffer,"%d",capacity->GetSelection());
   config->writeEntry(OPT_CAP,buffer);
   sprintf(buffer,"%d", fs->GetSelection());
   config->writeEntry(OPT_FS,buffer);
   sprintf(buffer,"%d", mkfsflag->GetValue());
   config->writeEntry(OPT_MKFS, buffer);
   sprintf(buffer,"%d", verifyflag->GetValue());
   config->writeEntry(OPT_VERIFY, buffer);

   rememberPosition();
   

   // Don't activate new language immediately or all sorts of strange
   // things happen
   //   ((wxExtApp *)wxTheApp)->SetLanguage(config->readEntry(OPT_LANG,OPT_LANG_D));
}	

static void GenericOk(wxButton& but, wxCommandEvent& event)
{	
   wxDialogBox *dialog = (wxDialogBox *)but.GetParent();
   char *help;
   
   dialog->Show(FALSE);
   dialog->Close();
   
   xfframe->setConfig();
   xfframe->SetSize(400,146);
}

void xfFrame::OptionsDialog(void)
{
   int dialogWidth = 350;
   int dialogHeight = 180;
   int nr;
   const char *help;

   dialog = new wxTabbedPanel(this, TRUE, 80, 100, 365, 180);
  
   wxButton *okButton = new wxButton(dialog, (wxFunction) GenericOk, _("Close"), 230, 80, 80, 25);
//   wxButton *cancelButton = new wxButton(dialog, NULL, _("Help"), 230, 130, 80, 25);

   // Note, omit the wxTAB_STYLE_COLOUR_INTERIOR, so we will guarantee a match
   // with the panel background, and save a bit of time.
   wxPanelTabView *view = new wxPanelTabView(dialog, wxTAB_STYLE_DRAW_BOX);
  
   wxRectangle rect;
   rect.x = 5;
   rect.y = 50;
   // Could calculate the view width from the tab width and spacing,
   // as below, but let's assume we have a fixed view width.
   //  rect.width = view->GetTabWidth()*4 + 3*view->GetHorizontalTabSpacing();
   rect.width = 326;
   rect.height = 180;
  
   view->SetViewRect(rect);

   // Calculate the tab width for 4 tabs, based on a view width of 326 and
   // the current horizontal spacing. Adjust the view width to exactly fit
   // the tabs.
   view->CalculateTabWidth(2, TRUE);

   if (!view->AddTab(OPT_TAB_DEV,        wxString(_("Device"))))
      return;
    
   if (!view->AddTab(OPT_TAB_FS,        wxString(_("Filesystem"))))
      return;

   if(!view->AddTab(OPT_TAB_VER,	wxString(_("Verify"))))
      return;

   if(!view->AddTab(OPT_TAB_LANG,	wxString(_("Other"))))
      return;

//   if(!view->AddTab(OPT_TAB_POS, wxString(_("Window"))))
//      return;

   config->setCurrentPath();
   config->setCurrentPath(APPNAME);
   // Add some panels
   wxPanel *panel1 = new wxPanel(dialog, rect.x + 20, rect.y + 10, 200, 200);

   for(nr = 0; drivenames[nr];nr++);
   drive =  new wxChoice(panel1,NULL,_("Device"), -1,-1,-1,-1, nr,drivenames);
   drive->SetSelection(atoi(config->readEntry(OPT_DEV,OPT_DEV_D)));
   panel1->NewLine();

   for(nr = 0; capacitynames[nr];nr++);
   capacity = new wxChoice(panel1, NULL,_("Capacity"),  -1,-1,-1,-1, nr, capacitynames);
   capacity->SetSelection(atoi(config->readEntry(OPT_CAP,OPT_CAP_D)));
   panel1->NewLine();

   view->AddTabPanel(OPT_TAB_DEV, panel1);

   
   wxPanel *panel2 = new wxPanel(dialog, rect.x + 20, rect.y + 10, 200, 250);
   panel2->SetLabelPosition(wxVERTICAL);

   mkfsflag = new wxCheckBox(panel2,NULL,_("Make Filesystem"));
   mkfsflag->SetValue(atoi(config->readEntry(OPT_MKFS,OPT_MKFS_D)));
   panel2->NewLine();
   
   for(nr = 0; filesystems[nr];nr++);
   fs = new wxChoice(panel2, NULL, _("Filesystem"), -1,-1,-1,-1, nr,
		     filesystems);
   fs->SetSelection(atoi(config->readEntry(OPT_FS,OPT_FS_D)));
   view->AddTabPanel(OPT_TAB_FS, panel2);
   
   wxPanel *panel3 = new wxPanel(dialog, rect.x + 20, rect.y + 10, 200, 250);
   panel3->SetLabelPosition(wxVERTICAL);

   verifyflag = new wxCheckBox(panel3,(wxFunction)NULL,_("Verify"));
   verifyflag->SetValue(atoi(config->readEntry(OPT_VERIFY,OPT_VERIFY_D)));
   
   view->AddTabPanel(OPT_TAB_VER, panel3);
  
   // The language panel:
   wxPanel *panel4 = new wxPanel(dialog, rect.x + 20, rect.y + 10, 200, 250);
   panel4->SetLabelPosition(wxVERTICAL);
   for(nr = 0; lang[nr];nr++);
   language =  new wxChoice(panel4,NULL,_("Language"), -1,-1,-1,-1, nr,lang);

   config->setCurrentPath("/");
   language->SetSelection(0);
   help = config->readEntry(OPT_LANG,OPT_LANG_D);
   if(help)
   {
      for(nr = 1; reallang[nr]; nr++)
	 if(strcmp(help,reallang[nr])==0)
	 {
	    language->SetSelection(nr);
	    break;
	 }
   }
   config->setCurrentPath(APPNAME);

   panel4->NewLine();
   positionflag = new wxCheckBox(panel4,(wxFunction)NULL,_("Remember window position"));
   positionflag->SetValue(atoi(config->readEntry(OPT_POSITION,OPT_POSITION_D)));
   
   view->AddTabPanel(OPT_TAB_LANG, panel4);


//   wxPanel *panel5 = new wxPanel(dialog, rect.x + 20, rect.y + 10, 200, 250);
//   panel5->SetLabelPosition(wxVERTICAL);

//   positionflag = new wxCheckBox(panel5,(wxFunction)NULL,_("Remember window position"));
//   positionflag->SetValue(atoi(config->readEntry(OPT_POSITION,OPT_POSITION_D)));
   
//   view->AddTabPanel(OPT_TAB_POS, panel5);

   
   // Don't know why this is necessary under Motif...
   dialog->SetSize(dialogWidth, dialogHeight-20);

   view->SetTabSelection(OPT_TAB_DEV);
  
//   Fit();
   dialog->Show(FALSE);  
}
